@@ -53,7 +53,7 @@
  *   transformed size of the #ClutterActor using the offscreen effect.
  *   Sub-classes of #ClutterOffscreenEffect can change the texture creation
  *   code to provide bigger textures by overriding the
- *   <function>create_target()</function> virtual function; no chain up
+ *   <function>create_texture()</function> virtual function; no chain up
  *   to the #ClutterOffscreenEffect implementation is required in this
  *   case.</para>
  * </refsect2>
@@ -75,7 +75,7 @@
 struct _ClutterOffscreenEffectPrivate
 {
   CoglHandle offscreen;
-  CoglHandle target;
+  CoglMaterial *target;
 
   ClutterActor *actor;
   ClutterActor *stage;
@@ -120,9 +120,9 @@ clutter_offscreen_effect_set_actor (ClutterActorMeta *meta,
 }
 
 static CoglHandle
-clutter_offscreen_effect_real_create_target (ClutterOffscreenEffect *effect,
-                                             gfloat                  width,
-                                             gfloat                  height)
+clutter_offscreen_effect_real_create_texture (ClutterOffscreenEffect *effect,
+                                              gfloat                  width,
+                                              gfloat                  height)
 {
   return cogl_texture_new_with_size (MAX (width, 1), MAX (height, 1),
                                      COGL_TEXTURE_NO_SLICING,
@@ -154,8 +154,8 @@ update_fbo (ClutterEffect *effect)
    * to make this work reliably
    */
   clutter_actor_get_transformed_size (priv->actor, &width, &height);
-  if (fabsf (priv->target_width - width) < 0.00001f &&
-      fabsf (priv->target_height - height) < 0.0001f)
+  if (priv->target_width == width &&
+      priv->target_height == height)
     return TRUE;
 
   if (priv->target != COGL_INVALID_HANDLE)
@@ -166,7 +166,7 @@ update_fbo (ClutterEffect *effect)
 
   priv->target = cogl_material_new ();
 
-  texture = clutter_offscreen_effect_create_target (self, width, height);
+  texture = clutter_offscreen_effect_create_texture (self, width, height);
   if (texture == COGL_INVALID_HANDLE)
      return FALSE;
 
@@ -174,7 +174,7 @@ update_fbo (ClutterEffect *effect)
   cogl_handle_unref (texture);
 
   /* we need to use the size of the texture target and not the minimum
-   * size we passed to the create_target() vfunc, as any sub-class might
+   * size we passed to the create_texture() vfunc, as any sub-class might
    * give use a bigger texture
    */
   priv->target_width = cogl_texture_get_width (texture);
@@ -250,7 +250,7 @@ clutter_offscreen_effect_pre_paint (ClutterEffect *effect)
 {
   ClutterOffscreenEffect *self = CLUTTER_OFFSCREEN_EFFECT (effect);
   ClutterOffscreenEffectPrivate *priv = self->priv;
-  ClutterPerspective perspective;
+  CoglMatrix projection;
   CoglColor transparent;
   CoglMatrix modelview;
   gfloat width, height;
@@ -265,15 +265,10 @@ clutter_offscreen_effect_pre_paint (ClutterEffect *effect)
     return FALSE;
 
   /* get the current modelview matrix so that we can copy it
-   * on the new framebuffer
+   * to the framebuffer
    */
   cogl_get_modelview_matrix (&modelview);
 
-  clutter_stage_get_perspective (CLUTTER_STAGE (priv->stage), &perspective);
-  clutter_actor_get_size (priv->stage, &width, &height);
-
-  get_screen_offsets (priv->actor, &priv->x_offset, &priv->y_offset);
-
   /* let's draw offscreen */
   cogl_push_framebuffer (priv->offscreen);
 
@@ -281,21 +276,25 @@ clutter_offscreen_effect_pre_paint (ClutterEffect *effect)
    * and it has its origin at the same position of the stage's; also
    * set up the perspective to be the same as the stage's
    */
+  clutter_actor_get_size (priv->stage, &width, &height);
+  get_screen_offsets (priv->actor, &priv->x_offset, &priv->y_offset);
   cogl_set_viewport (-priv->x_offset, -priv->y_offset, width, height);
-  cogl_perspective (perspective.fovy,
-                    perspective.aspect,
-                    perspective.z_near,
-                    perspective.z_far);
 
-  cogl_color_set_from_4ub (&transparent, 0, 0, 0, 0);
+  /* Copy the stage's projection matrix across to the framebuffer */
+  _clutter_stage_get_projection_matrix (CLUTTER_STAGE (priv->stage),
+                                        &projection);
+  cogl_set_projection_matrix (&projection);
+
+  /* Copy the modelview that would have been used if rendering onscreen */
+  cogl_set_modelview_matrix (&modelview);
+
+  cogl_color_init_from_4ub (&transparent, 0, 0, 0, 0);
   cogl_clear (&transparent,
               COGL_BUFFER_BIT_COLOR |
               COGL_BUFFER_BIT_DEPTH);
 
   cogl_push_matrix ();
 
-  cogl_set_modelview_matrix (&modelview);
-
   return TRUE;
 }
 
@@ -331,10 +330,7 @@ clutter_offscreen_effect_post_paint (ClutterEffect *effect)
 {
   ClutterOffscreenEffect *self = CLUTTER_OFFSCREEN_EFFECT (effect);
   ClutterOffscreenEffectPrivate *priv = self->priv;
-  ClutterPerspective perspective;
-  CoglMatrix modelview, matrix;
-  gfloat width, height;
-  gfloat z_camera;
+  CoglMatrix modelview;
 
   if (priv->offscreen == COGL_INVALID_HANDLE ||
       priv->target == COGL_INVALID_HANDLE ||
@@ -344,27 +340,16 @@ clutter_offscreen_effect_post_paint (ClutterEffect *effect)
   cogl_pop_matrix ();
   cogl_pop_framebuffer ();
 
-  clutter_stage_get_perspective (CLUTTER_STAGE (priv->stage), &perspective);
-  clutter_actor_get_size (priv->stage, &width, &height);
-
-  cogl_get_modelview_matrix (&modelview);
-  cogl_get_projection_matrix (&matrix);
-  z_camera = 0.5f * matrix.xx;
-
-  /* obliterate the current modelview matrix and reset it to be
-   * the same as the stage's at the beginning of a paint run; this
-   * is done to paint the target material in screen coordinates at
-   * the same place as the actor would have been
-   */
-  cogl_matrix_init_identity (&matrix);
-  cogl_matrix_translate (&matrix, -0.5f, -0.5f, -z_camera);
-  cogl_matrix_scale (&matrix, 1.0f / width, -1.0f / height, 1.0f / width);
-  cogl_matrix_translate (&matrix, 0.0f, -1.0f * height, 0.0f);
-  cogl_set_modelview_matrix (&matrix);
-
   cogl_push_matrix ();
 
-  cogl_translate (priv->x_offset, priv->y_offset, 0.0f);
+  /* Now reset the modelview to put us in stage coordinates so
+   * we can drawn the result of our offscreen render as a textured
+   * quad... */
+
+  cogl_matrix_init_identity (&modelview);
+  _clutter_actor_apply_modelview_transform (priv->stage, &modelview);
+  cogl_matrix_translate (&modelview, priv->x_offset, priv->y_offset, 0.0f);
+  cogl_set_modelview_matrix (&modelview);
 
   /* paint the target material; this is virtualized for
    * sub-classes that require special hand-holding
@@ -372,9 +357,6 @@ clutter_offscreen_effect_post_paint (ClutterEffect *effect)
   clutter_offscreen_effect_paint_target (self);
 
   cogl_pop_matrix ();
-
-  /* reset the modelview matrix */
-  cogl_set_modelview_matrix (&modelview);
 }
 
 static void
@@ -401,7 +383,7 @@ clutter_offscreen_effect_class_init (ClutterOffscreenEffectClass *klass)
 
   g_type_class_add_private (klass, sizeof (ClutterOffscreenEffectPrivate));
 
-  klass->create_target = clutter_offscreen_effect_real_create_target;
+  klass->create_texture = clutter_offscreen_effect_real_create_texture;
   klass->paint_target = clutter_offscreen_effect_real_paint_target;
 
   meta_class->set_actor = clutter_offscreen_effect_set_actor;
@@ -427,13 +409,16 @@ clutter_offscreen_effect_init (ClutterOffscreenEffect *self)
  * Retrieves the material used as a render target for the offscreen
  * buffer created by @effect
  *
- * Return value: (transfer none): a handle for a #CoglMaterial, or
- *   %COGL_INVALID_HANDLE. The returned handle is owned by Clutter
- *   and it should not be modified or freed
+ * You should only use the returned #CoglMaterial when painting. The
+ * returned material might change between different frames.
+ *
+ * Return value: (transfer none): a #CoglMaterial or %NULL. The
+ *   returned material is owned by Clutter and it should not be
+ *   modified or freed
  *
  * Since: 1.4
  */
-CoglHandle
+CoglMaterial *
 clutter_offscreen_effect_get_target (ClutterOffscreenEffect *effect)
 {
   g_return_val_if_fail (CLUTTER_IS_OFFSCREEN_EFFECT (effect),
@@ -459,26 +444,28 @@ clutter_offscreen_effect_paint_target (ClutterOffscreenEffect *effect)
 }
 
 /**
- * clutter_offscreen_effect_create_target:
+ * clutter_offscreen_effect_create_texture:
  * @effect: a #ClutterOffscreenEffect
  * @width: the minimum width of the target texture
  * @height: the minimum height of the target texture
  *
- * Calls the create_target() virtual function of the @effect
+ * Calls the create_texture() virtual function of the @effect
  *
- * Return value: a handle to the target texture
+ * Return value: (transfer full): a handle to a Cogl texture, or
+ *   %COGL_INVALID_HANDLE. The returned handle has its reference
+ *   count increased.
  *
  * Since: 1.4
  */
 CoglHandle
-clutter_offscreen_effect_create_target (ClutterOffscreenEffect *effect,
-                                        gfloat                  width,
-                                        gfloat                  height)
+clutter_offscreen_effect_create_texture (ClutterOffscreenEffect *effect,
+                                         gfloat                  width,
+                                         gfloat                  height)
 {
   g_return_val_if_fail (CLUTTER_IS_OFFSCREEN_EFFECT (effect),
                         COGL_INVALID_HANDLE);
 
-  return CLUTTER_OFFSCREEN_EFFECT_GET_CLASS (effect)->create_target (effect,
-                                                                     width,
-                                                                     height);
+  return CLUTTER_OFFSCREEN_EFFECT_GET_CLASS (effect)->create_texture (effect,
+                                                                      width,
+                                                                      height);
 }
